# 
#   Music API interface for Tcl
#
#   Author: Steve Landers <steve@DigitalSmarties.com>
#
#   Copyright (c) 2005 Eolas Technologies Inc
#   All rights reserved
#
#   The software is licensed under a Tcl compatible BSD style license, as
#   outlined at the end of this file
#
#   $Id: muzic.tcl,v 1.4 2005/08/15 09:24:46 steve Exp $

package provide muzic 1.0

if {[catch {package require critcl}]} {
    puts stderr "This extension requires Critcl to be compiled"
    exit 1
}

namespace eval ::muzic {

    namespace export init soundfont playnote channel

    # copy this file into the library so that any pure-Tcl procs are available
    # and also makes this source file part of the distribution
    critcl::tsources [file tail [info script]]

    # default to builtin sound (note on Linux this is a no-op)
    proc init {} {
        ::muzic::builtin_init
        ::muzic::soundfont builtin
        return
    }

    # soundfont both selects the soundfont, and switches between fluidsynth
    # and the builtin audio (if the font isn't specified)
    proc soundfont {{font ""}} {
        if {$font eq "" || $font eq "builtin"} {
            foreach cmd {noteon noteoff channel} {
                interp alias {} ::muzic::$cmd {} ::muzic::builtin_$cmd
            }
        } else {
            fluid_init
            foreach cmd {noteon noteoff channel} {
                interp alias {} ::muzic::$cmd {} ::muzic::fluid_$cmd
            }
            fluid_soundfont $font
        }
    }

    # play note using current audio driver
    proc playnote {chan pitch velocity {duration 500}} {
        variable type
        # normalize velocity to an integer in the range 0 to 100
        ::muzic::noteon $chan $pitch [expr {int($velocity*127/100)}]
        after $duration ::muzic::noteoff $chan $pitch
    }

    # path to Fluidsynth headers
    critcl::cflags -I./include

    # build fluidsynth support if possible
    if {![critcl::done] && [critcl::check "#include <fluidsynth.h>"]} {
        set fluidsynth 1
    } else {
        set fluidsynth 0
    }

    switch -glob [critcl::platform] {

        Darwin-* {
            # on OSX use the QuickTime Music APIs for the builtin driver

            variable builtin 1

            critcl::framework CoreAudio QuickTime Carbon

            critcl::ccode {
                #include <QuickTime/QuickTime.h>
                static NoteAllocator    na;
                static NoteChannel      channels[16];
            }

            proc fluid_init {} {
                do_fluid_init coreaudio
            }

            critcl::cproc builtin_init {} int {
                if (na = OpenDefaultComponent(kNoteAllocatorComponentType, 0))
                    return 1;
                return 0;
            }

            critcl::cproc builtin_noteon {int chan int pitch int velocity} int {
                if (chan < 0 || chan > 15)
                    return 0;
                NAPlayNote(na, channels[chan], pitch, velocity);
                return 1;
            }

            critcl::cproc builtin_noteoff {int chan int pitch} int {
                if (chan < 0 || chan > 15)
                    return 0;
                NAPlayNote(na, channels[chan], pitch, 0);
                return 1;
            }

            critcl::cproc builtin_channel {int chan int instrument} int {
                NoteRequest      nr;
                if (chan < 0 || chan > 15)
                    return 0;
                nr.info.flags = 0;
                nr.info.polyphony = 2;      // number of simultaneous tones
                nr.info.typicalPolyphony = 0x00010000;  // usually just one note
                if (NAStuffToneDescription(na, instrument, &nr.tone) != noErr) 
                    return 0;
                if (NANewNoteChannel(na, &nr, &channels[chan]) != noErr) {
                    CloseComponent(na);
                    return 0;
                }
                return 1;
            }

        }

        Linux-* {

            # we use Fluidsynth on Linux, which lacks a standard softsynth

            variable builtin 0
            
            if {![critcl::done] && !$fluidsynth} {
                error "fluidsynth is needed to build on Linux"
            }

            proc builtin_init {} {
                # no builtin softsynth on Linux
            }

            proc fluid_init {} {
                do_fluid_init oss
            }
                
            proc no_builtin {} {
                puts stderr \
                    "builtin sound not available on Linux - specify a soundfont"
            }

            proc builtin_noteon  {args} no_builtin
            proc builtin_noteoff {args} no_builtin
            proc builtin_channel {args} no_builtin

        }

        Windows-* {
            # on Windows use the MultiMedia SDK for the builtin driver

            variable builtin 1

            if {$::tcl_platform(os) eq "Windows 95"} {
                # builtin doesn't seem to work on Windows 95
                set builtin 0
            }

            if {$fluidsynth} {
                # we are going to statically link against libfluidsynth.a
                critcl::cflags -DFLUIDSYNTH_NOT_A_DLL
            } else {
                critcl::clibraries -lwinmm
            }

            critcl::ccode {
                #include <windows.h>
                #include <mmsystem.h>
                HMIDIOUT handle;

                static int shortmsg(int status, int data1, int data2) {
                    MMRESULT res;
                    res = midiOutShortMsg(handle,
                                    status | (data1 << 8) | (data2 << 16));
                    return (res == MMSYSERR_NOERROR ? 1 : 0);
                }
            }

            proc init {} {
                builtin_init
                soundfont
            }

            proc fluid_init {} {
                do_fluid_init dsound
            }

            critcl::cproc builtin_init {} int {
                MMRESULT res;
                res = midiOutOpen(&handle, MIDI_MAPPER, 0, 0, CALLBACK_NULL);
                return (res == MMSYSERR_NOERROR ? 1 : 0);
            }

            critcl::cproc builtin_noteon {int chan int pitch int velocity} int {
                return (shortmsg(chan + 144, pitch, velocity));
            }

            critcl::cproc builtin_noteoff {int chan int pitch} int {
                return (shortmsg(chan + 144, pitch, 0));
            }

            critcl::cproc builtin_channel {int chan int pitch} int {
                shortmsg(192 + chan, pitch, 1);
            }

        }

    }

    if {$fluidsynth} {

        # be careful to only include critcl procs in here, because this
        # block of code will only be run when building the package

        puts -nonewline stderr "(including fluidsynth support)"
        
    	# path to the fluidsynth library
    	critcl::clibraries -L[critcl::platform] -lfluidsynth

        switch -glob [critcl::platform] {
            Darwin-* {
                # make sure we pick up the library from our copy, in
                # case Fluidsynth is installed with different parameters
                critcl::ldflags -search_paths_first
            }
            Linux-* {
                critcl::clibraries -lpthread
            }
            Windows-* {
                # Fluidsynth uses DirectSound, which (in turn) uses winmm
                critcl::clibraries -ldsound -lwinmm
            } 
        }

        critcl::ccode {
            #include <fluidsynth.h>
            static fluid_synth_t *synth;
            static int sfont;
            static int initialised = 0;
        }

        critcl::cproc do_fluid_init {char* driver} int {
            if (!initialised) {
                fluid_settings_t     *settings;
                settings = new_fluid_settings();
                synth = new_fluid_synth(settings);
                fluid_settings_setstr(settings, "audio.driver", driver);
                if (new_fluid_audio_driver(settings, synth) == 0)
                    return 0;
                initialised = 1;
            }
            return 1;
        }

        critcl::cproc fluid_soundfont {char* path} int {
            return (sfont = fluid_synth_sfload(synth, path, 1));
        }
        
        critcl::cproc fluid_noteon {int chan int pitch int velocity} int {
            fluid_synth_noteon(synth, chan, pitch, velocity);
        }

        critcl::cproc fluid_noteoff {int chan int pitch} int {
            fluid_synth_noteoff(synth, chan, pitch);
        }

        critcl::cproc fluid_channel {int chan int instrument} int {
            fluid_synth_program_select(synth, chan, sfont, 0, instrument);
        }
    }

}

if 0 {

License Terms
-------------

This software is copyrighted Eolas Technologies Inc.
The following terms apply to all files associated with the software unless
explicitly disclaimed in individual files.

The authors hereby grant permission to use, copy, modify, distribute,
and license this software and its documentation for any purpose, provided
that existing copyright notices are retained in all copies and that this
notice is included verbatim in any distributions. No written agreement,
license, or royalty fee is required for any of the authorized uses.
Modifications to this software may be copyrighted by their authors
and need not follow the licensing terms described here, provided that
the new terms are clearly indicated on the first page of each file where
they apply.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

GOVERNMENT USE: If you are acquiring this software on behalf of the
U.S. government, the Government shall have only "Restricted Rights"
in the software and related documentation as defined in the Federal 
Acquisition Regulations (FARs) in Clause 52.227.19 (c) (2).  If you
are acquiring the software on behalf of the Department of Defense, the
software shall be classified as "Commercial Computer Software" and the
Government shall have only "Restricted Rights" as defined in Clause
252.227-7013 (c) (1) of DFARs.  Notwithstanding the foregoing, the
authors grant the U.S. Government and others acting in its behalf
permission to use and distribute the software in accordance with the
terms specified in this license. 

}
